"""
.. topic:: ``ih-github runner``

    A ``ih-github runner`` subcommand.

    See ``ih-github run --help`` for more details.
"""

import logging

import boto3
import click

from infrahouse_toolkit.cli.ih_github.cmd_runner.cmd_deregister import cmd_deregister
from infrahouse_toolkit.cli.ih_github.cmd_runner.cmd_download import cmd_download
from infrahouse_toolkit.cli.ih_github.cmd_runner.cmd_is_registered import (
    cmd_is_registered,
)
from infrahouse_toolkit.cli.ih_github.cmd_runner.cmd_list import cmd_list
from infrahouse_toolkit.cli.ih_github.cmd_runner.cmd_register import cmd_register
from infrahouse_toolkit.cli.ih_secrets.cmd_get import get_secret

LOG = logging.getLogger()


@click.group(
    name="runner",
)
@click.option("--github-token", help="Personal access token for GitHub.", envvar="GITHUB_TOKEN", show_default=True)
@click.option("--github-token-secret", help="Read GitHub token from AWS secret.")
@click.option("--registration-token-secret", help="AWS secret name with a registration token.", default=None)
@click.option(
    "--org",
    help="GitHub organization",
    required=False,
)
@click.pass_context
def cmd_runner(ctx, *args, **kwargs):
    """
    Manage self-hosted runners.

    """
    LOG.debug("args = %s", args)
    LOG.debug("kwargs = %s", kwargs)
    LOG.debug(ctx.args)
    if kwargs["github_token_secret"]:
        github_token = get_secret(boto3.client("secretsmanager"), kwargs["github_token_secret"])
    else:
        github_token = kwargs["github_token"]

    ctx.obj = {
        "github_token": github_token,
        "org": kwargs["org"],
        "registration_token_secret": kwargs["registration_token_secret"],
    }


for cmd in [cmd_list, cmd_register, cmd_deregister, cmd_is_registered, cmd_download]:
    # noinspection PyTypeChecker
    cmd_runner.add_command(cmd)



"""
.. topic:: ``ih-elastic cat shards``

    A ``ih-elastic cat shards`` subcommand.

    See ``ih-elastic cat shards --help`` for more details.
"""
import json
from logging import getLogger

import click
from elasticsearch.client import CatClient

LOG = getLogger(__name__)


@click.command(name="shards")
@click.pass_context
def cmd_shards(ctx):
    """
    Provides a detailed view of shard allocation on nodes.
    """
    client = CatClient(ctx.obj["es"])
    if ctx.obj["format"] == "json":
        print(json.dumps(client.shards(format=ctx.obj["format"], v=True).raw, indent=4))
    else:
        print(client.shards(format=ctx.obj["format"], v=True))



"""
.. topic:: ``ih-aws ecs``

    A ``ih-aws ecs`` subcommand.

    See ``ih-aws ecs --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_aws.cmd_ecs.cmd_wait_services_stable import (
    cmd_wait_services_stable,
)

LOG = getLogger(__name__)


@click.group(name="ecs")
def cmd_ecs():
    """
    AWS ECS Commands.
    """


for cmd in [
    cmd_wait_services_stable,
]:
    # noinspection PyTypeChecker
    cmd_ecs.add_command(cmd)



"""
.. topic:: ``ih-puppet``

    A group of commands to work with Puppet.

    See ``ih-puppet --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_puppet.cmd_apply import cmd_apply
from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


@click.group()
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--quiet",
    help="Suppress informational messages and output only warnings and errors.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--environment",
    help="Puppet environment",
    default="production",
    show_default=True,
)
@click.option(
    "--environmentpath",
    help="A path for directory environments.",
    default="{root_directory}/environments",
    show_default=True,
)
@click.option(
    "--root-directory",
    help="Path where the puppet code is hosted. "
    "The directory must include subdirectories ``environments``, ``modules``.",
    default="/opt/puppet-code",
    show_default=True,
)
@click.option(
    "--hiera-config",
    help="Path to hiera configuration file.",
    default="{root_directory}/environments/{environment}/hiera.yaml",
    show_default=True,
)
@click.option(
    "--module-path",
    help="Path to common puppet modules.",
    default="{root_directory}/modules",
    show_default=True,
)
@click.version_option()
@click.pass_context
def ih_puppet(ctx, **kwargs):
    """Puppet wrapper."""
    setup_logging(debug=kwargs["debug"], quiet=kwargs["quiet"])
    ctx.obj = {
        "debug": kwargs["debug"],
        "quiet": kwargs["quiet"],
        "environment": kwargs["environment"],
        "root_directory": kwargs["root_directory"],
        "hiera_config": kwargs["hiera_config"].format(
            root_directory=kwargs["root_directory"],
            environment=kwargs["environment"],
        ),
        "environmentpath": kwargs["environmentpath"].format(root_directory=kwargs["root_directory"]),
        "module_path": kwargs["module_path"].format(
            root_directory=kwargs["root_directory"],
        ),
    }


for cmd in [cmd_apply]:
    # noinspection PyTypeChecker
    ih_puppet.add_command(cmd)



"""
.. topic:: ``ih-s3-reprepro checkpool``

    A ``ih-s3-reprepro checkpool`` subcommand.

    See ``ih-s3-reprepro checkpool --help`` for more details.
"""

import click
from click import Context

from infrahouse_toolkit.cli.ih_s3_reprepro.utils import execute, local_s3


@click.command(name="checkpool")
@click.pass_context
def cmd_checkpool(ctx: Context):
    """Check if all files in the pool are still in proper shape."""
    bucket = ctx.parent.params["bucket"]
    role_arn = ctx.parent.params["role_arn"]
    with local_s3(bucket, role_arn, region=ctx.parent.params["aws_region"]) as path:
        execute(["reprepro", "-V", "-b", path, "checkpool"])



"""
.. topic:: ``ih-elastic cat``

    A ``ih-elastic cat`` subcommand.

    See ``ih-elastic cat --help`` for more details.
"""
from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_elastic.cmd_cluster.cmd_allocation_explain import (
    cmd_allocation_explain,
)

LOG = getLogger(__name__)


@click.group(name="cluster")
def cmd_cluster():
    """
    Cluster level operations.
    """


for cmd in [cmd_allocation_explain]:
    # noinspection PyTypeChecker
    cmd_cluster.add_command(cmd)



"""
.. topic:: ``utils.py``

    Various helper functions.
"""

import json
import os
import sys
from contextlib import contextmanager
from logging import getLogger
from os import getgid, getuid
from os import path as osp
from subprocess import CalledProcessError, Popen, check_call
from tempfile import TemporaryDirectory
from time import sleep, time

import boto3

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.cli.ih_s3_reprepro.aws import (
    assume_role,
    get_credentials_from_environ,
    get_credentials_from_profile,
)
from infrahouse_toolkit.cli.ih_s3_reprepro.gpg import gpg

DEPENDENCIES = ["reprepro", "gpg", "s3fs"]
LOG = getLogger()


def check_dependencies(binaries: list):
    """
    Ensure that dependencies are installed. The function calls each of the binary
    with a ``--help`` option.

    :param binaries: Dependency commands. List of strings.
    :type binaries: list
    """
    for dep in binaries:
        try:
            with open("/dev/null", "w", encoding=DEFAULT_OPEN_ENCODING) as devnull:
                LOG.debug("Checking if %s is installed.", dep)
                check_call([dep, "--help"], stdout=devnull, stderr=devnull)
        except FileNotFoundError:
            LOG.error("Looks like %s is not installed", dep)
            LOG.info("Try installing it by \n\n\tapt-get install %s\n", dep)
            sys.exit(1)


def mount_s3(bucket: str, path: str, role_arn: str = None, region: str = None):
    """
    Mount an S3 bucket at a path.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param path: Local filesystem path name.
    :type path: str
    :param role_arn: Assume role if specified.
    :param region: AWS region name.
    :type region: str
    """
    env = {}
    cmd = ["s3fs", bucket, path, "-o", f"uid={getuid()}", "-o", f"gid={getgid()}"]
    if role_arn:
        env = assume_role(role_arn)
        sts = boto3.client(
            "sts",
            aws_access_key_id=env["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=env["AWS_SECRET_ACCESS_KEY"],
            aws_session_token=env["AWS_SESSION_TOKEN"],
            region_name=region,
        )
        response = sts.get_caller_identity()
        LOG.debug("Assumed role: %s", response)
    elif "AWS_ACCESS_KEY_ID" in os.environ:
        env = get_credentials_from_environ()
    else:
        env = get_credentials_from_profile()

    LOG.debug("To reproduce environment: \n%s", "\n".join([f'export {key}="{value}"' for key, value in env.items()]))
    LOG.debug("Command to debug: mkdir -p %s; %s -o dbglevel=info -f -o curldbg", path, " ".join(cmd))
    execute(cmd, env=env)


def umount_s3(path: str):
    """
    Unmount an S3 bucket at a path.

    :param path: Local filesystem path name where the S3 bucket is mounted at.
    :type path: str
    """
    try:
        check_call(["umount", path])
    except CalledProcessError as err:
        LOG.exception(err)
        sys.exit(1)


@contextmanager
def local_s3(bucket, role_arn=None, retry_timeout=60, region=None) -> str:
    """
    Mount an S3 bucket locally and return a mount point.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param role_arn: Assume role if specified.
    :type role_arn: str
    :param retry_timeout: How many second to keep trying to mount the bucket.
    :type retry_timeout: int
    :param region: AWS region name.
    :type region: str
    :return: Local filesystem path where the S3 bucket is mounted at.
    """
    with TemporaryDirectory() as mnt_dir:
        try:
            now = time()
            timeout = now + retry_timeout
            while True:
                mount_s3(bucket, mnt_dir, role_arn=role_arn, region=region)
                if osp.exists(osp.join(mnt_dir, "conf/distributions")):
                    break
                LOG.warning("Waiting until s3://%s is mounted at %s", bucket, mnt_dir)
                sleep(1)
                if time() > timeout:
                    raise RuntimeError(f"s3://{bucket} is not mounted after {retry_timeout} seconds")
            yield mnt_dir
        except Exception as err:
            LOG.critical(err)
            raise

        finally:
            umount_s3(mnt_dir)


def execute(cmd: list, env: dict = None):
    """
    Execute a command and exit with 1 if the command raises CalledProcessError.

    :param cmd: A command to execute. It's passed to check_call() and therefore must be a list.
    :type cmd: list
    :param env: Pass a dictionary with environment
    :type env: dict
    """
    try:
        LOG.debug("Executing %s", " ".join(cmd))
        LOG.debug("Environment: %s", json.dumps(env, indent=4) if env else "None")
        with Popen(cmd, env=env) as proc:
            try:
                proc.communicate()
            except KeyboardInterrupt:
                LOG.info("Exiting on <ctrl>+c")
                proc.terminate()
                LOG.info("Process %s is terminated. Waiting for it to exit.", cmd[0])
                proc.wait(60)

    except CalledProcessError as err:
        LOG.exception(err)
        sys.exit(1)


@contextmanager
def repo_env(bucket, role_arn, gpg_key_secret_id, gpg_passphrase_secret_id, region=None):
    """
    Prepare locally a repo and GPG so "reprepro" can manage it.

    :param bucket: AWS S3 bucket with the repo. The repo must be in the root.
    :type bucket: str
    :param role_arn: Optional role ARN. If specified, AWS client will assume it.
    :type role_arn: str
    :param gpg_key_secret_id: AWS secretsmanager secret (name or ARN) that stores the private GPG key,
        needed by ``reprepro`` to sign the repo.
    :type gpg_key_secret_id: str
    :param gpg_passphrase_secret_id: AWS secretsmanager secret (name or ARN) that stores a passphrase for
        the private GPG key. Note, it's not the passphrase itself, it's a secret that stores it.
    :type gpg_passphrase_secret_id: str
    :param region: AWS region name.
    :type region: str
    :return: A tuple with two strings: Local filesystem directory with a mounted S3 bucket
        and GPG home directory.
    """
    with local_s3(bucket, role_arn, region=region) as path:
        with gpg(
            secret_key=gpg_key_secret_id, role_arn=role_arn, secret_passphrase=gpg_passphrase_secret_id, region=region
        ) as gpg_home:
            yield path, gpg_home



"""
Module for helper functions to deal with Terraform.
"""

import json
import os
import time
from base64 import b64decode
from contextlib import contextmanager
from logging import getLogger
from subprocess import PIPE, CalledProcessError, Popen

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.terraform.backends import get_backend
from infrahouse_toolkit.terraform.exceptions import IHParseError
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus, decolor

DEFAULT_PROGRESS_INTERVAL = 10
LOG = getLogger()


def parse_comment(comment_text: str) -> TFStatus:
    """
    Given a text comment try to instantiate a Terraform status out of it.
    Not any comment can yield a status, obviously.
    A comment with the status will have a metadata in its content::

        <details><summary><i>metadata</i></summary>

    If parser failed because the comment doesn't have the status
    then the function will raise an exception.

    :param comment_text: Comment text.
    :return: Terraform status.
    :rtype: TFStatus
    :raise IHParseError: When the comment text doesn't contain a status.
    """
    try:
        comment_as_lines = [line.strip() for line in comment_text.split("\n")]
        metadata_index = comment_as_lines.index("<details><summary><i>metadata</i></summary>")
        for idx in range(metadata_index, len(comment_as_lines)):
            if comment_as_lines[idx].startswith("```"):
                metadata = json.loads(b64decode(comment_as_lines[idx + 1].strip("`")))
                backend_url = list(metadata)[0]
                return TFStatus(
                    get_backend(backend_url),
                    success=metadata[backend_url]["success"],
                    run_result=RunResult(
                        metadata[backend_url]["add"],
                        metadata[backend_url]["change"],
                        metadata[backend_url]["destroy"],
                    ),
                    run_output=RunOutput(None, None),
                )
    except (ValueError, AttributeError) as err:
        raise IHParseError(f"Failed to parse comment: {comment_text}") from err

    raise IHParseError(f"Failed to parse comment: {comment_text}")


def parse_plan(output) -> (RunResult, RunResult):
    """
    Parse a string given by output and return a tuple with execution plan.

    Credit for a regexp removing colors from text:
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python

    :param output: Output of terraform plan command.
    :type output: str
    :return: Two tuples. One tuple with number of changes (add, change, destroy).
        Another - with list of to be added/changed/destroyed resources.
    :rtype: RunResult, RunResult
    """
    counts = RunResult(None, None, None)
    resources = RunResult(None, None, None)

    if output is None:
        return counts, resources

    output = decolor(output)

    if any(
        pattern in output
        for pattern in [
            "No changes. Infrastructure is up-to-date.",
            "No changes. Your infrastructure matches the configuration.",
        ]
    ):
        return RunResult(0, 0, 0), RunResult([], [], [])

    if "Terraform will perform the following actions" not in output:
        return counts, resources

    try:
        resources = RunResult([], [], [])
        for line in output.splitlines():
            if line.startswith("::debug::"):
                continue

            if line.startswith("Plan: "):
                split_line = line.split()
                # Plan: 4 to add, 11 to change, 7 to destroy.
                counts = RunResult(int(split_line[1]), int(split_line[4]), int(split_line[7]))
            elif "will be created" in line:
                resources.add.append(line.split()[1])
            elif "will be destroyed" in line:
                resources.destroy.append(line.split()[1])
            elif "will be updated in-place" in line:
                resources.change.append(line.split()[1])

    except AttributeError:
        pass

    return counts, resources


@contextmanager
def terraform_apply(
    path,
    destroy_after=True,
    json_output=False,
    var_file="terraform.tfvars",
    enable_trace=False,
):
    """
    Run terraform init and apply, then return a generator.
    If destroy_after is True, run terraform destroy afterward.

    :param path: Path to directory with terraform module.
    :type path: str
    :param destroy_after: Run terraform destroy after context it returned back.
    :type destroy_after: bool
    :param json_output: Yield terraform output result as a dict (available in the context)
    :type json_output: bool
    :param var_file: Path to a file with terraform variables.
    :type var_file: str
    :param enable_trace: If True, it will run ``terraform`` with ``TF_LOG=JSON`` and
        save the terraform trace in ``tf-apply-trace.txt`` and ``tf-destroy-trace.txt``.
        Useful if you want to find out what API calls terraform makes and for other
        debugging.
    :type enable_trace: bool
    :return: If json_output is true then yield the result from terraform_output otherwise nothing.
        Use it in the ``with`` block.
    :raise CalledProcessError: if either of terraform commands (except ``terraform destroy``)
        exits with non-zero.
    """
    cmds = [
        ["terraform", "init", "-no-color"],
        ["terraform", "get", "-update=true", "-no-color"],
        [
            "terraform",
            "apply",
            f"-var-file={var_file}",
            "-input=false",
            "-auto-approve",
        ],
    ]
    env = dict(os.environ)
    if enable_trace:
        env["TF_LOG"] = "JSON"
    try:
        for cmd in cmds:
            stderr = open("tf-apply-trace.txt", "w", encoding=DEFAULT_OPEN_ENCODING) if enable_trace else None
            ret, cout, cerr = execute(cmd, stdout=None, stderr=stderr, cwd=path, env=env)
            if ret:
                raise CalledProcessError(returncode=ret, cmd=" ".join(cmd), output=cout, stderr=cerr)
        if json_output:
            yield terraform_output(path)
        else:
            yield

    finally:
        if destroy_after:
            stderr = open("tf-destroy-trace.txt", "w", encoding=DEFAULT_OPEN_ENCODING) if enable_trace else None
            execute(
                [
                    "terraform",
                    "destroy",
                    f"-var-file={var_file}",
                    "-input=false",
                    "-auto-approve",
                ],
                stdout=None,
                stderr=stderr,
                cwd=path,
                env=env,
            )


def terraform_output(path):
    """
    Run terraform output and return the json results as a dict.

    :param path: Path to directory with terraform module.
    :type path: str
    :return: dict from terraform output
    :rtype: dict
    """
    cmd = ["terraform", "output", "-json"]
    ret, cout, cerr = execute(cmd, stdout=PIPE, stderr=None, cwd=path)
    if ret:
        raise CalledProcessError(returncode=ret, cmd=" ".join(cmd), output=cout, stderr=cerr)
    return json.loads(cout)


def execute(
    cmd,
    stdout=PIPE,
    stderr=PIPE,
    cwd=None,
    env=None,
):
    """
    Execute a command and return a tuple with return code, STDOUT and STDERR.

    :param cmd: Command.
    :type cmd: list
    :param stdout: Where to send stdout. Default PIPE.
    :type stdout: int, None
    :param stderr: Where to send stdout. Default PIPE.
    :type stderr: int, None
    :param cwd: Working directory.
    :type cwd: str
    :param env: Dictionary with environment for the process.
    :type env: dict
    :return: Tuple (return code, STDOUT, STDERR)
    :rtype: tuple
    """
    LOG.info("Executing: %s", " ".join(cmd))
    with Popen(cmd, stdout=stdout, stderr=stderr, cwd=cwd, env=env) as proc:
        last_checking = time.time()
        while True:
            if proc.poll() is not None:
                break
            if time.time() - last_checking > DEFAULT_PROGRESS_INTERVAL:
                LOG.info("Still waiting for process to complete.")
                last_checking = time.time()
            time.sleep(1)

        cout, cerr = proc.communicate()
        return proc.returncode, cout, cerr



"""
AWS classes.
"""

import sys
import time
import webbrowser
from logging import getLogger
from os import path as osp
from pprint import pformat
from time import sleep

import boto3
from boto3 import Session
from botocore.exceptions import (
    NoCredentialsError,
    SSOTokenLoadError,
    TokenRetrievalError,
)
from diskcache import Cache

from infrahouse_toolkit.aws.config import AWSConfig
from infrahouse_toolkit.aws.exceptions import IHAWSException
from infrahouse_toolkit.fs import ensure_permissions

LOG = getLogger()


def aws_sso_login(aws_config: AWSConfig, profile_name: str, region: str = None):
    """
    Login into AWS using SSO.

    Credit:
    https://stackoverflow.com/questions/62311866/how-to-use-the-aws-python-sdk-while-connecting-via-sso-credentials
    """
    cache_directory = osp.expanduser("~/.infrahouse-toolkit")
    with Cache(directory=cache_directory) as cache_reference:
        cache_key = f"ih-ec2-credentials-{profile_name}"
        credentials = cache_reference.get(cache_key)
        if not credentials:
            credentials = _get_credentials(aws_config, profile_name)
            cache_reference.set(
                cache_key, credentials, expire=int(int(credentials["expiration"]) / 1000 - int(time.time()))
            )

    ensure_permissions(cache_directory, 0o700)
    return Session(
        region_name=region or aws_config.get_region(profile_name),
        aws_access_key_id=credentials["accessKeyId"],
        aws_secret_access_key=credentials["secretAccessKey"],
        aws_session_token=credentials["sessionToken"],
    )


def get_aws_client(service_name: str, profile: str, region: str, session=None):
    """
    Get a client instance for an AWS service.

    :param service_name: AWS service e.g. ``ec2``.
    :param profile: AWS profile for authentication.
    :param region: AWS region.
    :param session: if an AWS session is passed, use it to create a client.
    :type session: Session
    :return: A client instance.
    """
    session = session or Session(region_name=region, profile_name=profile)
    return session.client(service_name)


def get_aws_session(aws_config: AWSConfig, aws_profile: str, aws_region: str) -> Session:
    """

    :param aws_config:
    :param aws_profile:
    :param aws_region:
    :return: Authenticated AWS session, or None if boto3 can connect to AWS without extra steps.
    """
    if aws_profile is None and "default" in aws_config.profiles:
        aws_profile = "default"

    try:
        response = get_aws_client("sts", aws_profile, aws_region).get_caller_identity()
        LOG.info("Connected to AWS as %s", response["Arn"])

    except (SSOTokenLoadError, TokenRetrievalError) as err:
        if not aws_profile:
            LOG.error("Try to run ih-aws with --aws-profile option.")
            LOG.error("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
            sys.exit(1)
        LOG.debug(err)
        aws_session = aws_sso_login(aws_config, aws_profile, region=aws_region)
        response = get_aws_client("sts", aws_profile, aws_region, session=aws_session).get_caller_identity()
        LOG.info("Connected to AWS as %s", response["Arn"])
        return aws_session

    except NoCredentialsError as err:
        LOG.error(err)
        LOG.info("Try to run ih-aws with --aws-profile option.")
        LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)

    return boto3.Session(region_name=aws_region)


def _get_credentials(aws_config: AWSConfig, profile_name: str):
    """
    Login into AWS using SSO.

    Credit:
    https://stackoverflow.com/questions/62311866/how-to-use-the-aws-python-sdk-while-connecting-via-sso-credentials

    :raise IHAWSException: If user didn't confirm auth
    """

    session = Session()
    sso_oidc = session.client("sso-oidc")
    client_creds = sso_oidc.register_client(
        clientName="infrahouse-toolkit",
        clientType="public",
    )
    LOG.debug("client_creds = %s", pformat(client_creds, indent=4))
    device_authorization = sso_oidc.start_device_authorization(
        clientId=client_creds["clientId"],
        clientSecret=client_creds["clientSecret"],
        startUrl=aws_config.get_start_url(profile_name),
    )

    LOG.debug("device_authorization = %s", pformat(device_authorization, indent=4))
    url = device_authorization["verificationUriComplete"]
    device_code = device_authorization["deviceCode"]
    expires_in = device_authorization["expiresIn"]
    interval = device_authorization["interval"]
    LOG.info("Verify user code: %s", device_authorization["userCode"])
    webbrowser.open(url, autoraise=True)
    for _ in range(1, expires_in // interval + 1):
        sleep(interval)
        try:
            token = sso_oidc.create_token(
                grantType="urn:ietf:params:oauth:grant-type:device_code",
                deviceCode=device_code,
                clientId=client_creds["clientId"],
                clientSecret=client_creds["clientSecret"],
            )
            access_token = token["accessToken"]
            sso = session.client("sso")
            role_creds = sso.get_role_credentials(
                roleName=aws_config.get_role(profile_name),
                accountId=aws_config.get_account_id(profile_name),
                accessToken=access_token,
            )["roleCredentials"]
            return role_creds
        except sso_oidc.exceptions.AuthorizationPendingException:
            pass

    raise IHAWSException(f"The verification code isn't confirmed by user in {expires_in} seconds")



"""
.. topic:: ``ih-plan publish``

    A ``ih-plan publish`` subcommand.

    See ``ih-plan publish --help`` for more details.
"""

import click

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.cli.lib import get_backend_key, get_bucket
from infrahouse_toolkit.terraform import RunOutput, TFStatus, parse_plan
from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.githubpr import GitHubPR
from infrahouse_toolkit.terraform.status import strip_lines


@click.command(name="publish")
@click.option("--github-token", help="Personal access token for GitHub.", envvar="GITHUB_TOKEN")
@click.option(
    "--tf-exit-code", help="With what code number the terraform plan command exited.", default=0, show_default=True
)
@click.option(
    "--private-gist/--public-gist",
    help="When the comment is too large and the tool needs to publish the comment as a gist,"
    "should it be private or public.",
    default=True,
    show_default=True,
)
@click.argument("repo")
@click.argument("pull_request_number")
@click.argument("tf_plan_stdout", type=click.Path(exists=True))
@click.argument("tf_plan_stderr", type=click.Path(exists=True))
@click.pass_context
def cmd_publish(*args, **kwargs):
    """
    Publish Terraform plan to GitHub pull request.

    Example:

        ih-plan publish infrahouse8/github-control 33 plan.stdout plan.stderr

    Here:

    \b
        * ``infrahouse8/github-control`` - full repository name.
        * ``33`` - pull request number.
        * ``plan.stdout`` - file with ``terraform plan`` output.
        * ``plan.stderr`` - file with ``terraform plan`` error output.

    """
    ctx = args[0]

    with open(kwargs["tf_plan_stdout"], encoding=DEFAULT_OPEN_ENCODING) as fp_stdout, open(
        kwargs["tf_plan_stderr"], encoding=DEFAULT_OPEN_ENCODING
    ) as fp_stderr:
        stdout = strip_lines(fp_stdout.read(), "::debug::")
        stderr = strip_lines(fp_stderr.read(), "::debug::")

        counts, resources = parse_plan(stdout)
        backend = TFS3Backend(
            ctx.obj["bucket"] or get_bucket(ctx.obj["tf_backend_file"]),
            get_backend_key(ctx.obj["tf_backend_file"]),
        )
        status = TFStatus(
            backend,
            kwargs["tf_exit_code"] == 0,
            counts,
            RunOutput(stdout, stderr),
            affected_resources=resources,
        )
        pull_request = GitHubPR(kwargs["repo"], int(kwargs["pull_request_number"]), github_token=kwargs["github_token"])
        comment = pull_request.find_comment_by_backend(backend)
        if comment:
            pull_request.edit_comment(comment, status.comment, private_gist=kwargs["private_gist"])
        else:
            pull_request.publish_comment(status.comment, private_gist=kwargs["private_gist"])



"""
Module for :py:class:`GitHubPR`.
"""

from logging import getLogger
from os import environ
from typing import Union

from github import Github, InputFileContent
from github.GithubException import GithubException
from github.IssueComment import IssueComment

from infrahouse_toolkit.terraform import IHParseError, parse_comment
from infrahouse_toolkit.terraform.backends.tfbackend import TFBackend

LOG = getLogger()


class GitHubPR:
    """
    :py:class:`GitHubPR` represents a pull request on GitHub.
    The pull request is identified by a repository name and a pull request number.
    To access GitHub the class needs a GitHub token.
    All these are input arguments of the class.

    :param repo_name: Full repository name. For example, ``infrahouse/infrahouse-toolkit``.
    :typo repo_name: str
    :param pull_request: Pull request number.
    :type pull_request: int
    :param github_token: GitHub personal access tokens. They are created in
        https://github.com/settings/tokens
    :type github_token: str
    """

    def __init__(self, repo_name: str, pull_request: int, github_token: str = None):
        self._github_token = github_token
        self._repo_name = repo_name
        self._pr_number = pull_request

    @property
    def comments(self):
        """
        An interator with comments in this PR.
        """
        return self.pull_request.get_issue_comments()

    @property
    def github(self):
        """
        GitHub client.
        """
        return Github(login_or_token=self.github_token)

    @property
    def github_token(self):
        """
        GitHub token as passed by the class argument or from the ``GITHUB_TOKEN`` environment variable.
        If the ``GITHUB_TOKEN`` environment variable is not defined the property will return None.
        """
        return self._github_token if self._github_token else environ.get("GITHUB_TOKEN")

    @property
    def repo(self):
        """
        Repository object of the repository name passed in the class argument.
        """
        return self.github.get_repo(self._repo_name)

    @property
    def pull_request(self):
        """
        Pull request object of the repository name passed in the class argument.
        """
        return self.repo.get_pull(self._pr_number)

    def delete_my_comments(self):
        """
        Delete all comments in the pull request.
        """
        for comment in self.comments:
            comment.delete()

    def find_comment_by_backend(self, backend: TFBackend) -> Union[IssueComment, None]:
        """
        Find a comment that describes state of a given backend.
        It will return None if nothing is found.

        :param backend: Terraform Backend configuration.
        :return: a comment object or None.
        :rtype: IssueComment, None
        """
        for comment in self.comments:
            try:
                status = parse_comment(comment.body)
                if status.backend == backend:
                    return comment
            except IHParseError:
                pass
        return None

    def edit_comment(self, comment: IssueComment, new_text: str, private_gist: bool = True):
        """
        Modify existing comment. If the new comment is too big,
        publish it as a gist.
        """
        try:
            comment.edit(new_text)
        except GithubException as err:
            LOG.error(err)
            # https://docs.github.com/en/rest/issues/comments?apiVersion=2022-11-28#create-an-issue-comment
            if err.status == 422:  # Validation failed, or the endpoint has been spammed.
                gist = self._publish_gist(
                    f"pr-{self._pr_number}-plan",
                    f"{self._repo_name.replace('/', '-')}-pr-{self._pr_number}-plan.txt",
                    comment,
                    not private_gist,
                )
                comment.edit(f"Comment was too big. It's published as a gist at {gist.html_url}.")
            else:
                raise

    def publish_comment(self, comment: str, private_gist: bool = True):
        """Add the given text as a comment in the pull request."""
        try:
            self.pull_request.create_issue_comment(comment)
        except GithubException as err:
            LOG.error(err)
            # https://docs.github.com/en/rest/issues/comments?apiVersion=2022-11-28#create-an-issue-comment
            if err.status == 422:  # Validation failed, or the endpoint has been spammed.
                gist = self._publish_gist(
                    f"pr-{self._pr_number}-plan",
                    f"{self._repo_name.replace('/', '-')}-pr-{self._pr_number}-plan.txt",
                    comment,
                    not private_gist,
                )
                self.pull_request.create_issue_comment(
                    f"Comment was too big. It's published as a gist at {gist.html_url}."
                )
            else:
                raise

    def _publish_gist(self, gist_id, filename, content, public):
        current_user = self.github.get_user()
        return current_user.create_gist(
            public=public,
            files={
                gist_id: InputFileContent(
                    content=content,
                    new_name=filename,
                )
            },
        )



"""
.. topic:: ``ih-elastic cat snapshots``

    A ``ih-elastic cat snapshots`` subcommand.

    See ``ih-elastic cat snapshots --help`` for more details.
"""
from logging import getLogger

import click
from elasticsearch.client import CatClient

LOG = getLogger(__name__)


@click.command(name="snapshots")
@click.pass_context
def cmd_snapshots(ctx):
    """
    Returns all snapshots.
    """
    client = CatClient(ctx.obj["es"])
    LOG.info("\n%s", client.snapshots(repository="_all", format=ctx.obj["format"], v=True))



"""
.. topic:: ``ih-github run``

    A ``ih-github run`` subcommand.

    See ``ih-github run --help`` for more details.
"""

import logging
import sys
import time
import traceback
from subprocess import PIPE, Popen

import click

from infrahouse_toolkit.terraform.githubpr import GitHubPR

LOG = logging.getLogger()


@click.command(
    name="run",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@click.option("--github-token", help="Personal access token for GitHub.", envvar="GITHUB_TOKEN")
@click.option("--run-timeout", help="How many seconds the command it allowed to run", show_default=True, default=3600)
@click.argument("repo")
@click.argument("pull_request_number", type=click.INT)
@click.pass_context
def cmd_run(ctx, *args, **kwargs):
    """
    Run a command and publish its output to as a comment in GitHub pull request.

    For instance
    """
    LOG.debug("args = %s", args)
    LOG.debug("kwargs = %s", kwargs)
    LOG.debug(ctx.args)
    cmd = ctx.args
    comment = None
    pull_request = GitHubPR(kwargs["repo"], kwargs["pull_request_number"], github_token=kwargs["github_token"])
    try:
        with Popen(cmd, stderr=PIPE, stdout=PIPE) as proc:
            LOG.info("Launched command: %s", " ".join(cmd))

            start_time = time.time()
            stop_time = start_time + kwargs["run_timeout"]
            return_code = None
            wait_time = 1
            while time.time() < stop_time:
                return_code = proc.poll()
                if return_code is None:
                    if time.time() - start_time > 1:
                        LOG.info("%s is still running. Waiting %d seconds more.", cmd[0], stop_time - time.time())
                    wait_time = wait_time * 2 if wait_time < 10 else 10
                    time.sleep(wait_time)
                else:
                    break

            cout, cerr = proc.communicate()

            comment = f"Command `{' '.join(cmd)}` exited with code {return_code}.\n"
            comment += f"**Standard output**:\n```\n{cout.decode() or 'No output'}\n```\n"
            comment += f"**Standard error**:\n```\n{cerr.decode() or 'No output'}\n```\n"

            sys.stdout.write(cout.decode())
            sys.stderr.write(cerr.decode())
            sys.exit(proc.returncode)

    except FileNotFoundError as err:
        LOG.exception(err)
        comment = f"Command `{' '.join(cmd)}` failed to start.\n"
        comment += f"Stack trace:\n```\n{traceback.format_exc()}\n```\n"
        sys.exit(1)

    finally:
        pull_request.publish_comment(comment)



"""
.. topic:: ``ih-elastic snapshots status``

    A ``ih-elastic snapshots status`` subcommand.

    See ``ih-elastic snapshots status --help`` for more details.
"""

import json
from logging import getLogger

import click
from elasticsearch.client import SnapshotClient

LOG = getLogger(__name__)


@click.command(name="status")
@click.argument("repository")
@click.argument("snapshot")
@click.pass_context
def cmd_status(ctx, **kwargs):
    """
    Returns information about the status of a snapshot.
    """
    client = SnapshotClient(ctx.obj["es"])
    print(json.dumps(client.status(repository=kwargs["repository"], snapshot=kwargs["snapshot"]).body, indent=4))



"""
.. topic:: ``ih-elastic cat repositories``

    A ``ih-elastic cat repositories`` subcommand.

    See ``ih-elastic cat repositories --help`` for more details.
"""
from logging import getLogger

import click
from elasticsearch.client import CatClient

LOG = getLogger(__name__)


@click.command(name="repositories")
@click.pass_context
def cmd_repositories(ctx):
    """
    Returns information about snapshot repositories registered in the cluster.
    """
    client = CatClient(ctx.obj["es"])
    LOG.info("\n%s", client.repositories(format=ctx.obj["format"], v=True))



"""
.. topic:: ``ih-ec2 instance-types``

    A ``ih-ec2 instance-types`` subcommand.

    See ``ih-ec2 instance-types`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import ClientError
from tabulate import tabulate

LOG = getLogger()


def list_instance_types(ec2_client):
    """
    Find and print supported instance types in a region.
    """

    def _format_storage_info(i_type):
        if field in i_type:
            disks = []
            for disk in i_type[field]["Disks"]:
                disks.append(f"{disk['Type']}: {disk['Count']} * {disk['SizeInGB']} GB")
            return f"{i_type[field]['TotalSizeInGB'] } GB: {' '.join(disks)}"
        return ""

    instance_types = []
    header = ["InstanceType", "ProcessorInfo", "MemoryInfo", "InstanceStorageInfo", "NetworkInfo"]
    next_token = None
    while True:
        kwargs = {}
        if next_token:
            kwargs["NextToken"] = next_token
        response = ec2_client.describe_instance_types(**kwargs)

        for instance_type in response["InstanceTypes"]:
            row = []
            for field in header:
                if field == "ProcessorInfo":
                    architectures = instance_type[field]["SupportedArchitectures"]
                    speed = (
                        instance_type[field]["SustainedClockSpeedInGhz"]
                        if "SustainedClockSpeedInGhz" in instance_type[field]
                        else ""
                    )
                    value = f"{', '.join(architectures)} {speed} GHz"

                elif field == "MemoryInfo":
                    value = f"{round(instance_type[field]['SizeInMiB'] / 1024) } GB"

                elif field == "InstanceStorageInfo":
                    value = _format_storage_info(instance_type)
                elif field == "NetworkInfo":
                    value = instance_type[field]["NetworkPerformance"]
                else:
                    value = instance_type[field] if field in instance_type else ""
                row.append(value)
            instance_types.append(row)

        if "NextToken" in response:
            next_token = response["NextToken"]
        else:
            break

    print(tabulate(sorted(instance_types, key=lambda x: x[0]), headers=header, tablefmt="outline"))


@click.command(name="instance-types")
@click.pass_context
def cmd_instance_types(ctx):
    """
    Describe AWS EC2 instance types.
    """
    ec2_client = ctx.obj["ec2_client"]
    try:
        list_instance_types(ec2_client)
    except ClientError as err:
        LOG.exception(err)
        sys.exit(1)



"""
.. topic:: ``ih-ec2 terminate``

    A ``ih-ec2 terminate`` subcommand.

    See ``ih-ec2 terminate`` for more details.
"""

import sys
from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_ec2.cmd_list import list_ec2_instances

LOG = getLogger()


def terminate_ec2_instance(ec2_client, instance_id):
    """Terminate an EC2 instance."""
    ec2_client.terminate_instances(
        InstanceIds=[
            instance_id,
        ],
    )
    LOG.info("Successfully terminated %s", instance_id)


@click.command(name="terminate")
@click.argument("instance_id", required=False)
@click.pass_context
def cmd_terminate(ctx, instance_id):
    """
    Terminate an EC2 instance.
    """
    ec2_client = ctx.obj["ec2_client"]
    if not instance_id:
        LOG.error("Please specify INSTANCE_ID from following:")
        list_ec2_instances(ec2_client)
        sys.exit(1)

    if click.prompt(f"Are you sure you want to terminate {instance_id}? (yes/no)"):
        terminate_ec2_instance(ctx.obj["ec2_client"], instance_id)



from textwrap import dedent

import pytest

from infrahouse_toolkit.terraform.status import strip_lines


@pytest.mark.parametrize(
    "in_text, pattern, out_text",
    [
        ("foo", "bar", "foo"),
        ("", "bar", ""),
        (
            dedent(
                """
                test line 1
                test line 2
                test line 3
                ::debug::Terraform exited with code 0.
                ::debug::stdout: module.jumphost
                ::debug::stderr:
                ::debug::exitcode: 0
                """
            ),
            "::debug::",
            dedent(
                """
                test line 1
                test line 2
                test line 3
                """
            ),
        ),
    ],
)
def test_strip_lines(in_text, pattern, out_text):
    assert strip_lines(in_text, pattern) == out_text



"""
.. topic:: ``ih-s3-reprepro remove``

    A ``ih-s3-reprepro remove`` subcommand.

    See ``ih-s3-reprepro remove --help`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import BotoCoreError, ClientError
from click import Context

from infrahouse_toolkit.cli.ih_s3_reprepro.utils import execute, repo_env

LOG = getLogger()


@click.command(name="remove")
@click.argument("codename")
@click.argument("package_name")
@click.pass_context
def cmd_remove(ctx: Context, codename, package_name):
    """Delete  all packages in the specified distribution, that have package name listed as argument."""
    try:
        with repo_env(
            ctx.parent.params["bucket"],
            ctx.parent.params["role_arn"],
            ctx.parent.params["gpg_key_secret_id"],
            ctx.parent.params["gpg_passphrase_secret_id"],
            region=ctx.parent.params["aws_region"],
        ) as (path, gpg_home):
            execute(["reprepro", "-V", "-b", path, "--gnupghome", gpg_home, "remove", codename, package_name])

    except (ClientError, BotoCoreError) as err:
        LOG.error(err)
        sys.exit(1)



"""
.. topic:: ``ih-s3-reprepro list``

    A ``ih-s3-reprepro list`` subcommand.

    See ``ih-s3-reprepro list --help`` for more details.
"""

import click
from click import Context

from infrahouse_toolkit.cli.ih_s3_reprepro.utils import execute, local_s3


@click.command(name="list")
@click.argument("distribution")
@click.argument("package_name", nargs=-1)
@click.pass_context
def cmd_list(ctx: Context, distribution: str, package_name: str):
    """List all packages by the given name occurring in the given distribution."""
    bucket = ctx.parent.params["bucket"]
    role_arn = ctx.parent.params["role_arn"]
    with local_s3(bucket, role_arn, region=ctx.parent.params["aws_region"]) as path:
        cmd = ["reprepro", "-V", "-b", path, "list", distribution]
        cmd.extend(package_name)
        execute(cmd)



"""Auxiliary functions for command line tools."""

import json

import boto3
import click
import hcl

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING

DEFAULT_TF_BACKEND_FILE = "terraform.tf"


def get_bucket(tf_file=DEFAULT_TF_BACKEND_FILE) -> str:
    """
    Find bucket name in a Terraform backend configuration.

    :param tf_file: Path to the Terraform backend configuration.
    :type tf_file: str
    :return: Bucket name.
    """
    with open(tf_file, encoding=DEFAULT_OPEN_ENCODING) as f_desc:
        obj = hcl.load(f_desc)
        return obj["terraform"]["backend"]["s3"]["bucket"]


def get_backend_key(tf_file=DEFAULT_TF_BACKEND_FILE) -> str:
    """
    Find terraform state filename in a Terraform backend configuration.

    :param tf_file: Path to the Terraform backend configuration.
    :type tf_file: str
    :return: Path to Terraform state in S3.
    """
    with open(tf_file, encoding=DEFAULT_OPEN_ENCODING) as f_desc:
        obj = hcl.load(f_desc)
        return obj["terraform"]["backend"]["s3"]["key"]


def get_s3_client(role: str = None):
    """
    Get a boto3 S3 client to work with AWS S3.
    If a role is given, assume it.

    :param role: ARN of a role to be assumed
    :return: A boto3 S3 client object
    """
    if role:
        client = boto3.client("sts")
        response = client.assume_role(DurationSeconds=900, RoleArn=role, RoleSessionName="infrahouse-toolkit")
        session = boto3.Session(
            aws_access_key_id=response["Credentials"]["AccessKeyId"],
            aws_secret_access_key=response["Credentials"]["SecretAccessKey"],
            aws_session_token=response["Credentials"]["SessionToken"],
        )
    else:
        session = boto3.Session()

    return session.client("s3")


def get_elastic_password(secret_key="elastic_secret"):
    """
    Try to extract the password for user elastic from AWS secretsmanager.

    If the code runs on an elasticsearch node, there is a secret-id with the password in the custom facts.
    Try to extract that secret and return the password.

    :param secret_key: A key in the puppet facts map facts["elasticsearch"][<secret_key>].
        ``elastic_secret`` or ``kibana_system_secret`` are the only supported values.
    :type secret_key: str
    """
    try:
        with open("/etc/puppetlabs/facter/facts.d/custom.json", encoding=DEFAULT_OPEN_ENCODING) as f_custom_facts:
            custom_facts = json.load(f_custom_facts)
            client = boto3.client("secretsmanager")
            response = client.get_secret_value(SecretId=custom_facts["elasticsearch"][secret_key])
            return response["SecretString"]

    except FileNotFoundError:
        return None


def read_from_file_or_prompt(file_path: str, prompt_text="Enter a secret value and press ENTER") -> str:
    """
    Read a string from a file if it exists. If not, prompt a user to enter the string.
    Return the string value.

    :param file_path: Path to the file.
    :type file_path: str
    :param prompt_text: What text to show a user.
    :type prompt_text: str
    :return: The string value whether it was read from the file or entered by teh user.
    """
    if file_path:
        with open(file_path, encoding=DEFAULT_OPEN_ENCODING) as val_desc:
            return val_desc.read()
    else:
        return click.prompt(prompt_text, hide_input=True)



"""
InfraHouse Toolkit file system module.
"""

from logging import getLogger
from os import chmod, stat

LOG = getLogger(__name__)


def ensure_permissions(path: str, permissions: int):
    """
    For a path on the file system check permissions and set if they differ.

    :param path: Filesystem path to a file or directory.
    :type path: str
    :param permissions: Permissions the file or directory must have. Can be an integer like 0o755 or 0o644.
    :type permissions: int
    """
    result = stat(path)
    set_permissions = result.st_mode & 0o777
    LOG.debug("%s permissions: 0o%o", path, set_permissions)
    if set_permissions != permissions:
        LOG.debug("Setting %s permissions to 0o%o", path, permissions)
        chmod(path, permissions)



import json
from base64 import b64decode
from os import path as osp

import pytest

from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


@pytest.mark.parametrize(
    "plan_file, result_counts, expected_comment",
    [
        (
            "plan-no-output.stdout",
            (1, 1, 1),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|  Success  |   🟢 Add |   🟡 Change |   🔴 Destroy |
|:---------:|--------:|-----------:|------------:|
|     ✅     |       1 |          1 |           1 |
<details>
<summary>STDOUT</summary>

```
no stdout
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAxLCAiY2hhbmdlIjogMSwgImRlc3Ryb3kiOiAxfX0=
```
</details>""",
        ),
        (
            "plan-0-0-0.stdout",
            (0, 0, 0),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|  Success  |    Add |    Change |    Destroy |
|:---------:|-------:|----------:|-----------:|
|     ✅     |      0 |         0 |          0 |
<details>
<summary>STDOUT</summary>

```
Terraform has compared your real infrastructure against your configuration
and found no differences, so no changes are needed.
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAwLCAiY2hhbmdlIjogMCwgImRlc3Ryb3kiOiAwfX0=
```
</details>""",
        ),
        (
            "plan-2-0-0.stdout",
            (2, 0, 0),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|  Success  |   🟢 Add |    Change |    Destroy |
|:---------:|--------:|----------:|-----------:|
|     ✅     |       2 |         0 |          0 |
<details>
<summary>STDOUT</summary>

```
Terraform used the selected providers to generate the following execution
plan. Resource actions are indicated with the following symbols:
  + create

Terraform will perform the following actions:

  # module.repos["test"].github_repository.repo will be created
  + resource "github_repository" "repo" {
      + allow_auto_merge            = false
      + allow_merge_commit          = true
      + allow_rebase_merge          = true
      + allow_squash_merge          = true
      + archived                    = false
      + default_branch              = (known after apply)
      + delete_branch_on_merge      = false
      + description                 = "Template for a GitHub Control repository"
      + etag                        = (known after apply)
      + full_name                   = (known after apply)
      + git_clone_url               = (known after apply)
      + has_issues                  = true
      + html_url                    = (known after apply)
      + http_clone_url              = (known after apply)
      + id                          = (known after apply)
      + merge_commit_message        = "PR_TITLE"
      + merge_commit_title          = "MERGE_MESSAGE"
      + name                        = "test"
      + node_id                     = (known after apply)
      + private                     = (known after apply)
      + repo_id                     = (known after apply)
      + squash_merge_commit_message = "COMMIT_MESSAGES"
      + squash_merge_commit_title   = "COMMIT_OR_PR_TITLE"
      + ssh_clone_url               = (known after apply)
      + svn_url                     = (known after apply)
      + visibility                  = "public"

      + security_and_analysis {
          + advanced_security {
              + status = (known after apply)
            }

          + secret_scanning {
              + status = (known after apply)
            }

          + secret_scanning_push_protection {
              + status = (known after apply)
            }
        }
    }

  # module.repos["test"].github_team_repository.dev will be created
  + resource "github_team_repository" "dev" {
      + etag       = (known after apply)
      + id         = (known after apply)
      + permission = "push"
      + repository = "test"
      + team_id    = "7332815"
    }

Plan: 2 to add, 0 to change, 0 to destroy.

─────────────────────────────────────────────────────────────────────────────

Saved the plan to: tf.plan

To perform exactly these actions, run the following command to apply:
    terraform apply "tf.plan"
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAyLCAiY2hhbmdlIjogMCwgImRlc3Ryb3kiOiAwfX0=
```
</details>""",
        ),
        (
            "plan-2-1-2.stdout",
            (2, 1, 2),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|  Success  |   🟢 Add |   🟡 Change |   🔴 Destroy |
|:---------:|--------:|-----------:|------------:|
|     ✅     |       2 |          1 |           2 |
<details>
<summary>STDOUT</summary>

```
Terraform used the selected providers to generate the following execution
plan. Resource actions are indicated with the following symbols:
  + create
  ~ update in-place
  - destroy

Terraform will perform the following actions:

  # module.repos["cookiecutter-github-control"].github_repository.repo will be destroyed
  # (because module.repos["cookiecutter-github-control"] is not in configuration)
  - resource "github_repository" "repo" {
      - allow_auto_merge            = false -> null
      - allow_merge_commit          = true -> null
      - allow_rebase_merge          = true -> null
      - allow_squash_merge          = true -> null
      - allow_update_branch         = false -> null
      - archived                    = false -> null
      - default_branch              = "main" -> null
      - delete_branch_on_merge      = false -> null
      - description                 = "Template for a GitHub Control repository" -> null
      - etag                        = "W/\\\"8b4a792bc1474d381caaa63b76668993b4adc42ae76ed53d6886d8562ebb0c67\\\"" -> null
      - full_name                   = "infrahouse/cookiecutter-github-control" -> null
      - git_clone_url               = "git://github.com/infrahouse/cookiecutter-github-control.git" -> null
      - has_discussions             = false -> null
      - has_downloads               = false -> null
      - has_issues                  = true -> null
      - has_projects                = false -> null
      - has_wiki                    = false -> null
      - html_url                    = "https://github.com/infrahouse/cookiecutter-github-control" -> null
      - http_clone_url              = "https://github.com/infrahouse/cookiecutter-github-control.git" -> null
      - id                          = "cookiecutter-github-control" -> null
      - is_template                 = false -> null
      - merge_commit_message        = "PR_TITLE" -> null
      - merge_commit_title          = "MERGE_MESSAGE" -> null
      - name                        = "cookiecutter-github-control" -> null
      - node_id                     = "R_kgDOI528zg" -> null
      - private                     = false -> null
      - repo_id                     = 597540046 -> null
      - squash_merge_commit_message = "COMMIT_MESSAGES" -> null
      - squash_merge_commit_title   = "COMMIT_OR_PR_TITLE" -> null
      - ssh_clone_url               = "git@github.com:infrahouse/cookiecutter-github-control.git" -> null
      - svn_url                     = "https://github.com/infrahouse/cookiecutter-github-control" -> null
      - topics                      = [] -> null
      - visibility                  = "public" -> null
      - vulnerability_alerts        = false -> null

      - security_and_analysis {

          - secret_scanning {
              - status = "disabled" -> null
            }

          - secret_scanning_push_protection {
              - status = "disabled" -> null
            }
        }
    }

  # module.repos["cookiecutter-github-control"].github_team_repository.dev will be destroyed
  # (because module.repos["cookiecutter-github-control"] is not in configuration)
  - resource "github_team_repository" "dev" {
      - etag       = "W/\\\"8043f81b19693f6c1a72d21bb8dc03859c98bf78bfbe79782bfe13fa813992ca\\\"" -> null
      - id         = "7332815:cookiecutter-github-control" -> null
      - permission = "push" -> null
      - repository = "cookiecutter-github-control" -> null
      - team_id    = "7332815" -> null
    }

  # module.repos["infrahouse-toolkit"].github_repository.repo will be updated in-place
  ~ resource "github_repository" "repo" {
      ~ description                 = "InfraHouse Toolkit" -> "InfraHouse Toolkit1"
        id                          = "infrahouse-toolkit"
        name                        = "infrahouse-toolkit"
        # (31 unchanged attributes hidden)

        # (1 unchanged block hidden)
    }

  # module.repos["test"].github_repository.repo will be created
  + resource "github_repository" "repo" {
      + allow_auto_merge            = false
      + allow_merge_commit          = true
      + allow_rebase_merge          = true
      + allow_squash_merge          = true
      + archived                    = false
      + default_branch              = (known after apply)
      + delete_branch_on_merge      = false
      + description                 = "Template for a GitHub Control repository"
      + etag                        = (known after apply)
      + full_name                   = (known after apply)
      + git_clone_url               = (known after apply)
      + has_issues                  = true
      + html_url                    = (known after apply)
      + http_clone_url              = (known after apply)
      + id                          = (known after apply)
      + merge_commit_message        = "PR_TITLE"
      + merge_commit_title          = "MERGE_MESSAGE"
      + name                        = "test"
      + node_id                     = (known after apply)
      + private                     = (known after apply)
      + repo_id                     = (known after apply)
      + squash_merge_commit_message = "COMMIT_MESSAGES"
      + squash_merge_commit_title   = "COMMIT_OR_PR_TITLE"
      + ssh_clone_url               = (known after apply)
      + svn_url                     = (known after apply)
      + visibility                  = "public"

      + security_and_analysis {
          + advanced_security {
              + status = (known after apply)
            }

          + secret_scanning {
              + status = (known after apply)
            }

          + secret_scanning_push_protection {
              + status = (known after apply)
            }
        }
    }

  # module.repos["test"].github_team_repository.dev will be created
  + resource "github_team_repository" "dev" {
      + etag       = (known after apply)
      + id         = (known after apply)
      + permission = "push"
      + repository = "test"
      + team_id    = "7332815"
    }

Plan: 2 to add, 1 to change, 2 to destroy.

─────────────────────────────────────────────────────────────────────────────

Saved the plan to: tf.plan

To perform exactly these actions, run the following command to apply:
    terraform apply "tf.plan"
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAyLCAiY2hhbmdlIjogMSwgImRlc3Ryb3kiOiAyfX0=
```
</details>""",
        ),
    ],
)
def test_comment(plan_file, result_counts, expected_comment):
    with open(osp.join("infrahouse_toolkit/terraform/tests/plans", plan_file)) as fp:
        status = TFStatus(
            TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(*result_counts), RunOutput(fp.read(), None)
        )
        assert isinstance(status.comment, str)
        print("\nActual comment:")
        print(status.comment)
        print("EOF Actual comment.")
        assert status.comment == expected_comment


def test_comment_none():
    status = TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"),
        True,
        RunResult(None, None, None),
        RunOutput("no stdout", "no stderr"),
        affected_resources=RunResult(None, None, None),
    )
    assert isinstance(status.comment, str)
    print(status.comment)



"""
.. topic:: ``ih-elastic cat shards``

    A ``ih-elastic cat shards`` subcommand.

    See ``ih-elastic cat shards --help`` for more details.
"""
import json
import sys
from logging import getLogger

import click
from elasticsearch import BadRequestError
from elasticsearch.client import ClusterClient

LOG = getLogger(__name__)


@click.command(name="allocation-explain")
@click.option(
    "--index",
    help="Specifies the name of the index that you would like an explanation for."
    " Run ih-elastic cat shards to get a list.",
    default=None,
    show_default=True,
)
@click.option(
    "--shard",
    help="Specifies the ID of the shard that you would like an explanation for."
    " Run ih-elastic cat shards to get a list.",
    default=None,
    show_default=True,
    type=click.INT,
)
@click.option(
    "--primary/--replica",
    help="If --primary, returns explanation for the primary shard for the given shard ID.",
    default=None,
    show_default=True,
    is_flag=True,
)
@click.pass_context
def cmd_allocation_explain(ctx, **kwargs):
    """
    Provides explanations for shard allocations in the cluster.
    """
    client = ClusterClient(ctx.obj["es"])
    try:
        ae_kwargs = {}
        for arg in ["index", "shard"]:
            if kwargs[arg]:
                ae_kwargs[arg] = kwargs[arg]
        if kwargs["primary"] is not None:
            ae_kwargs["primary"] = kwargs["primary"]

        LOG.info(json.dumps(client.allocation_explain(**ae_kwargs).body, indent=4))
    except BadRequestError as err:
        LOG.error(json.dumps(err.body, indent=4))
        sys.exit(1)
